from typing import Optional
from warnings import warn

import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class PeakletPositionsBase(strax.Plugin):
    """Base class for reconstructions.

    This class should only be used when subclassed for the different algorithms. Provides
    x_algorithm, y_algorithm for all peaklets > than min_reconstruction_area based on the top array.

    """

    __version__ = "0.0.0"

    depends_on = "peaklets"
    algorithm: Optional[str] = None
    compressor = "zstd"

    min_reconstruction_area = straxen.URLConfig(
        help="Skip reconstruction if area (PE) is less than this",
        default=0,
        infer_type=False,
    )

    n_top_pmts = straxen.URLConfig(
        default=straxen.n_top_pmts, infer_type=False, help="Number of top PMTs"
    )

    fix_steps_per_execution = straxen.URLConfig(
        track=False,
        default=True,
        infer_type=False,
        help=(
            "Fix steps_per_execution to 1 for the model, "
            "this is a patch when keras and tensorflow are not fully compatible"
        ),
    )

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(
                f"Base class should not be used without algorithm as done in {__class__.__name__}"
            )
        dtype = [
            (
                "x_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S2 X position (cm), uncorrected",
            ),
            (
                "y_" + self.algorithm,
                np.float32,
                f"Reconstructed {self.algorithm} S2 Y position (cm), uncorrected",
            ),
        ]
        dtype += strax.time_fields
        return dtype

    def get_tf_model(self):
        """Simple wrapper to have several tf_model_mlp, tf_model_cnf, ..

        point to this same function in the compute method

        """
        model = getattr(self, f"tf_model_{self.algorithm}", None)
        if model is None:
            warn(
                f"Setting model to None for {self.__class__.__name__} will "
                f"set only nans as output for {self.algorithm}"
            )
        else:
            if self.fix_steps_per_execution and hasattr(model, "steps_per_execution"):
                if getattr(model, "steps_per_execution", None) is None:
                    model.steps_per_execution = 1
        if isinstance(model, str):
            raise ValueError(
                f"open files from tf:// protocol! Got {model} "
                "instead, see tests/test_posrec.py for examples."
            )
        return model

    def compute(self, peaklets):
        result = np.ones(len(peaklets), dtype=self.dtype)
        result["time"], result["endtime"] = peaklets["time"], strax.endtime(peaklets)

        result["x_" + self.algorithm] *= np.nan
        result["y_" + self.algorithm] *= np.nan
        model = self.get_tf_model()

        if model is None:
            # This plugin is disabled since no model is provided
            return result

        # Keep large peaklets only
        peak_mask = peaklets["area"] > self.min_reconstruction_area
        if not np.sum(peak_mask):
            # Nothing to do, and .predict crashes on empty arrays
            return result

        # Getting actual position reconstruction
        area_per_channel_top = peaklets["area_per_channel"][peak_mask, : self.n_top_pmts]
        with np.errstate(divide="ignore", invalid="ignore"):
            area_per_channel_top = area_per_channel_top / np.max(
                area_per_channel_top, axis=1
            ).reshape(-1, 1)
        area_per_channel_top = area_per_channel_top.reshape(-1, self.n_top_pmts)
        output = model.predict(area_per_channel_top, verbose=0)

        # writing output to the result
        result["x_" + self.algorithm][peak_mask] = output[:, 0]
        result["y_" + self.algorithm][peak_mask] = output[:, 1]
        return result



import strax
from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.peaks.peak_basics_vanilla import PeakBasicsVanilla


export, __all__ = strax.exporter()


@export
class PeakBasicsHighEnergy(PeakBasicsVanilla):
    __doc__ = HE_PREAMBLE + (PeakBasicsVanilla.__doc__ or "")

    __version__ = "0.0.2"
    depends_on = "peaks_he"
    provides = "peak_basics_he"
    child_ends_with = "_he"

    def compute(self, peaks_he):
        return super().compute(peaks_he)



import strax
import straxen
from straxen.plugins.peaks._peak_s1_positions_base import PeakS1PositionBase


export, __all__ = strax.exporter()


@export
class PeakS1PositionCNN(PeakS1PositionBase):
    """S1 CNN for (x,y,z) position S1 reconstruction at peak level."""

    provides = "peak_s1_positions_cnn"
    algorithm = "s1_cnn"
    __version__ = "0.0.1"

    tf_model_s1_cnn = straxen.URLConfig(
        default=(
            "tf://"
            "resource://"
            "xedocs://posrec_models"
            "?version=ONLINE"
            "&run_id=plugin.run_id"
            "&kind=s1_cnn"
            "&fmt=abs_path"
            "&attr=value"
        ),
        help=(
            's1 position 3d reconstruction cnn model. Should be opened using the "tf" descriptor. '
            'Set to "None" to skip computation'
        ),
        cache=3,
    )



from datetime import datetime

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pytz
import strax
import straxen

export, __all__ = strax.exporter()

# Default attributes to display in the event_display (looks little
# complicated but just repeats same fields for S1 S1)
# Should be of form as below where {v} wil be filled with the value of
# event['key']: (('key', '{v} UNIT'), ..)
PEAK_DISPLAY_DEFAULT_INFO: list = sum(
    [
        [
            (k.format(i=s_i), u)
            for k, u in (
                ("cs{i}", "{v:.2f} PE"),
                ("s{i}_area", "{v:.2f} PE"),
                ("alt_cs{i}", "{v:.2f} PE"),
                ("s{i}_n_channels", "{v}"),
                ("s{i}_area_fraction_top", "{v:.2f}"),
                ("s{i}_range_50p_area", "{v:.1f}"),
            )
        ]
        for s_i in (1, 2)
    ],
    [],
)
EVENT_DISPLAY_DEFAULT_INFO = (
    ("time", "{v} ns"),
    ("endtime", "{v} ns"),
    ("event_number", "{v}"),
    ("x", "{v:.2f} cm"),
    ("y", "{v:.2f} cm"),
    ("z", "{v:.2f} cm"),
    ("r", "{v:.2f} cm"),
    ("theta", "{v:.2f} rad"),
    ("drift_time", "{v} ns"),
    ("alt_s1_interaction_drift_time", "{v} ns"),
    ("alt_s2_interaction_drift_time", "{v} ns"),
)


# Don't be smart with the arguments, since it is a minianalyses we
# need to have all the arguments
@straxen.mini_analysis(requires=("event_info",))
def event_display(
    context,
    run_id,
    events,
    to_pe,
    records_matrix=True,
    s2_fuzz=50,
    s1_fuzz=0,
    max_peaks=500,
    s1_hp_kwargs=None,
    s2_hp_kwargs=None,
    event_time_limit=None,
    plot_all_positions=True,
    display_peak_info=PEAK_DISPLAY_DEFAULT_INFO,
    display_event_info=EVENT_DISPLAY_DEFAULT_INFO,
    simple_layout=False,
):
    """{event_docs} {event_returns}"""
    if records_matrix not in ("raw", True, False):
        raise ValueError('Choose either "raw", True or False for records_matrix')
    if (records_matrix == "raw" and not context.is_stored(run_id, "raw_records")) or (
        isinstance(records_matrix, bool) and not context.is_stored(run_id, "records")
    ):  # noqa
        print("(raw)records not stored! Not showing records_matrix")
        records_matrix = False
    # Convert string to int to allow plots to be enlarged for extra panel
    _rr_resize_int = int(bool(records_matrix))

    if simple_layout:
        axes = _event_display_simple_layout()
    else:
        axes = _event_display_full_layout(_rr_resize_int, records_matrix)

    return _event_display(
        context,
        run_id,
        events,
        to_pe,
        axes=axes,
        records_matrix=records_matrix,
        s2_fuzz=s2_fuzz,
        s1_fuzz=s1_fuzz,
        max_peaks=max_peaks,
        display_peak_info=display_peak_info,
        display_event_info=display_event_info,
        s1_hp_kwargs=s1_hp_kwargs,
        s2_hp_kwargs=s2_hp_kwargs,
        event_time_limit=event_time_limit,
        plot_all_positions=plot_all_positions,
    )


def _event_display(
    context,
    run_id,
    events,
    to_pe,
    axes=None,
    records_matrix=True,
    s2_fuzz=50,
    s1_fuzz=0,
    max_peaks=500,
    display_peak_info=PEAK_DISPLAY_DEFAULT_INFO,
    display_event_info=EVENT_DISPLAY_DEFAULT_INFO,
    s1_hp_kwargs=None,
    s2_hp_kwargs=None,
    event_time_limit=None,
    plot_all_positions=True,
):
    """{event_docs}

    :param axes: if a dict of matplotlib axes (w/ same keys as below, and empty/None for panels not
        filled) {event_returns}

    """
    if len(events) != 1:
        raise ValueError(f"Found {len(events)} only request one")
    event = events[0]

    if axes is None:
        raise ValueError(f"No axes provided")
    ax_s1 = axes.get("ax_s1", None)
    ax_s2 = axes.get("ax_s2", None)
    ax_s1_hp_t = axes.get("ax_s1_hp_t", None)
    ax_s1_hp_b = axes.get("ax_s1_hp_b", None)
    ax_s2_hp_t = axes.get("ax_s2_hp_t", None)
    ax_s2_hp_b = axes.get("ax_s2_hp_b", None)
    ax_event_info = axes.get("ax_event_info", None)
    ax_peak_info = axes.get("ax_peak_info", None)
    ax_ev = axes.get("ax_ev", None)
    ax_rec = axes.get("ax_rec", None)

    # titles
    for ax, title in zip(
        [ax_s1, ax_s1_hp_t, ax_s1_hp_b, ax_s2, ax_s2_hp_t, ax_s2_hp_b, ax_event_info, ax_peak_info],
        [
            "Main S1",
            "S1 top",
            "S1 bottom",
            "Main S2",
            "S2 top",
            "S2 bottom",
            "Event info",
            "Peak info",
        ],
    ):
        if ax is not None:
            ax.set_title(title)

    # Parse the hit pattern options
    # Convert to dict (not at function definition because of mutable defaults)
    if s1_hp_kwargs is None:
        s1_hp_kwargs = {}
    if s2_hp_kwargs is None:
        s2_hp_kwargs = {}

    # Hit patterns options:
    for hp_opt, color_map in ((s1_hp_kwargs, "Blues"), (s2_hp_kwargs, "Greens")):
        _common_opt = dict(
            pmt_label_color="lightgrey",
            log_scale=True,
            vmin=0.1,
            s=(250 if records_matrix else 220),
            pmt_label_size=7,
            edgecolor="grey",
            dead_pmts=np.argwhere(to_pe == 0),
            cmap=color_map,
        )
        # update s1 & S2 hit pattern kwargs with _common_opt if not
        # specified by the user
        for k, v in _common_opt.items():
            if k not in hp_opt:
                hp_opt[k] = v

    # S1
    if events["s1_area"] != 0:
        if ax_s1 is not None:
            plt.sca(ax_s1)
            context.plot_peaks(
                run_id,
                time_range=(events["s1_time"] - s1_fuzz, events["s1_endtime"] + s1_fuzz),
                single_figure=False,
            )

        # Hit pattern plots
        area = context.get_array(
            run_id,
            "peaklets",
            time_range=(events["s1_time"], events["s1_endtime"]),
            keep_columns=("area_per_channel", "time", "dt", "length"),
            progress_bar=False,
        )
        for ax, array in ((ax_s1_hp_t, "top"), (ax_s1_hp_b, "bottom")):
            if ax is not None:
                plt.sca(ax)
                straxen.plot_on_single_pmt_array(
                    c=np.sum(area["area_per_channel"], axis=0), array_name=array, **s1_hp_kwargs
                )
                # Mark reconstructed position
                plt.scatter(event["x"], event["y"], marker="X", s=100, c="k")

    # S2
    if event["s2_area"] != 0:
        if ax_s2 is not None:
            plt.sca(ax_s2)
            context.plot_peaks(
                run_id,
                time_range=(events["s2_time"] - s2_fuzz, events["s2_endtime"] + s2_fuzz),
                single_figure=False,
            )

        # Hit pattern plots
        area = context.get_array(
            run_id,
            "peaklets",
            time_range=(events["s2_time"], events["s2_endtime"]),
            keep_columns=("area_per_channel", "time", "dt", "length"),
            progress_bar=False,
        )
        for axi, (ax, array) in enumerate([(ax_s2_hp_t, "top"), (ax_s2_hp_b, "bottom")]):
            if ax is not None:
                plt.sca(ax)
                straxen.plot_on_single_pmt_array(
                    c=np.sum(area["area_per_channel"], axis=0), array_name=array, **s2_hp_kwargs
                )
                # Mark reconstructed position (corrected)
                plt.scatter(event["x"], event["y"], marker="X", s=100, c="k")
                if axi == 0 and plot_all_positions:
                    _scatter_rec(event)

    # Fill panels with peak/event info
    for it, (ax, labels_and_unit) in enumerate(
        [(ax_event_info, display_event_info), (ax_peak_info, display_peak_info)]
    ):
        if ax is not None:
            for i, (_lab, _unit) in enumerate(labels_and_unit):
                coord = 0.01, 0.9 - 0.9 * i / len(labels_and_unit)
                ax.text(*coord, _lab[:24], va="top", zorder=-10)
                ax.text(coord[0] + 0.5, coord[1], _unit.format(v=event[_lab]), va="top", zorder=-10)
                # Remove axes and labels from panel
                ax.set_xticks([])
                ax.set_yticks([])
                _ = [s.set_visible(False) for s in ax.spines.values()]

    # Plot peaks in event
    ev_range = None
    if ax_ev is not None:
        plt.sca(ax_ev)
        if event_time_limit is None:
            time_range = (events["time"], events["endtime"])
        else:
            time_range = event_time_limit

        context.plot_peaks(
            run_id, time_range=time_range, show_largest=max_peaks, single_figure=False
        )
        ev_range = plt.xlim()

    if records_matrix and ax_rec is not None:
        plt.sca(ax_rec)
        context.plot_records_matrix(
            run_id,
            raw=records_matrix == "raw",
            time_range=(events["time"], events["endtime"]),
            single_figure=False,
        )
        ax_rec.tick_params(axis="x", rotation=0)
        # Top vs bottom division
        ax_rec.axhline(straxen.n_top_pmts, c="k")
        if ev_range is not None:
            plt.xlim(*ev_range)

    # Final tweaks
    if ax_s2 is not None:
        ax_s1.tick_params(axis="x", rotation=45)
    if ax_s2 is not None:
        ax_s1.tick_params(axis="x", rotation=45)
    if ax_ev is not None:
        ax_ev.tick_params(axis="x", rotation=0)
    title = (
        f"Run {run_id}. Time "
        f"{str(events['time'])[:-9]}.{str(events['time'])[-9:]}\n"
        f"{datetime.fromtimestamp(event['time'] / 1e9, tz=pytz.utc)}"
    )
    plt.suptitle(title, y=0.95)
    # NB: reflects panels order
    return (
        ax_s1,
        ax_s2,
        ax_s1_hp_t,
        ax_s1_hp_b,
        ax_event_info,
        ax_peak_info,
        ax_s2_hp_t,
        ax_s2_hp_b,
        ax_ev,
        ax_rec,
    )


@straxen.mini_analysis(requires=("event_info",))
def event_display_simple(context, run_id, events, **kwargs):
    raise NotImplementedError("Pass st.event_display(.., simple_layout=True)")


def _event_display_simple_layout() -> dict:
    """Setup a simple gidspec for the event display."""
    fig = plt.figure(figsize=(12, 8), facecolor="white")
    grid = plt.GridSpec(2, 3, hspace=0.5)
    axes = dict()
    axes["ax_s1"] = fig.add_subplot(grid[0, 0])
    axes["ax_s2"] = fig.add_subplot(grid[0, 1])
    axes["ax_s2_hp_t"] = fig.add_subplot(grid[0, 2])
    axes["ax_ev"] = fig.add_subplot(grid[1, :])
    return axes


def _event_display_full_layout(_rr_resize_int, records_matrix) -> dict:
    """Setup the full gidspec for the event display."""
    fig = plt.figure(figsize=(25, 21 if _rr_resize_int else 16), facecolor="white")
    grid = plt.GridSpec(
        (2 + _rr_resize_int),
        1,
        hspace=0.1 + 0.1 * _rr_resize_int,
        height_ratios=[1.5, 0.5, 0.5][: 2 + _rr_resize_int],
    )

    # S1, S2, hitpatterns
    gss_0 = gridspec.GridSpecFromSubplotSpec(2, 4, subplot_spec=grid[0], wspace=0.25, hspace=0.4)
    ax_s1 = fig.add_subplot(gss_0[0])
    ax_s2 = fig.add_subplot(gss_0[1])
    ax_s1_hp_t = fig.add_subplot(gss_0[2])
    ax_s1_hp_b = fig.add_subplot(gss_0[3])
    ax_s2_hp_t = fig.add_subplot(gss_0[6])
    ax_s2_hp_b = fig.add_subplot(gss_0[7])

    # Peak & event info
    ax_event_info = fig.add_subplot(gss_0[4])
    ax_peak_info = fig.add_subplot(gss_0[5])

    # All peaks in event
    gss_1 = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=grid[1])
    ax_ev = fig.add_subplot(gss_1[0])
    ax_rec = None

    # (raw)records matrix (optional)
    if records_matrix:
        gss_2 = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=grid[2])
        ax_rec = fig.add_subplot(gss_2[0])
    axes = dict(
        ax_s1=ax_s1,
        ax_s2=ax_s2,
        ax_s1_hp_t=ax_s1_hp_t,
        ax_s1_hp_b=ax_s1_hp_b,
        ax_event_info=ax_event_info,
        ax_peak_info=ax_peak_info,
        ax_s2_hp_t=ax_s2_hp_t,
        ax_s2_hp_b=ax_s2_hp_b,
        ax_ev=ax_ev,
        ax_rec=ax_rec,
    )
    return axes


@export
def plot_single_event(context: strax.Context, run_id, events, event_number=None, **kwargs):
    """Wrapper for event_display.

    :param context: strax.context
    :param run_id: run id
    :param events: dataframe / numpy array of events. Should either be length 1 or the event_number
        argument should be provided
    :param event_number: (optional) int, if provided, only show this event number
    :param kwargs: kwargs for events_display
    :return: see events_display

    """
    if event_number is not None:
        events = events[events["event_number"] == event_number]
    if len(events) > 1 or len(events) == 0:
        raise ValueError(
            f"Make sure to provide an event number or a single event. Got {len(events)} events"
        )

    return context.event_display(
        run_id, time_range=(events[0]["time"], events[0]["endtime"]), **kwargs
    )


def _scatter_rec(
    _event,
    recs=None,
    scatter_kwargs=None,
):
    """Convenient wrapper to show posrec of three algorithms for xenonnt."""
    if recs is None:
        recs = ("mlp", "cnf")
    elif len(recs) > 5:
        raise ValueError("I only got five markers/colors")
    if scatter_kwargs is None:
        scatter_kwargs = {}
    scatter_kwargs.setdefault("s", 100)
    scatter_kwargs.setdefault("alpha", 0.8)
    shapes = ("v", "^", ">", "<", "*", "D", "P")
    colors = ("brown", "orange", "lightcoral", "gold", "lime", "crimson")
    for _i, _r in enumerate(recs):
        x, y = _event[f"s2_x_{_r}"], _event[f"s2_y_{_r}"]
        if np.isnan(x) or np.isnan(y):
            continue
        plt.scatter(
            x,
            y,
            marker=shapes[_i],
            c=colors[_i],
            label=_r.upper(),
            **scatter_kwargs,
        )
    plt.legend(loc="best", fontsize="x-small", markerscale=0.5)


# Event display docstrings.
# Let's add them to the corresponding functions

event_docs = """
Make a waveform-display of a given event. Requires events, peaks and
    peaklets (optionally: records). NB: time selection should return
    only one event!

:param context: strax.Context provided by the minianalysis wrapper
:param run_id: run-id of the event
:param events: events, provided by the minianalysis wrapper
:param to_pe: gains, provided by the minianalysis wrapper
:param records_matrix: False (no record matrix), True, or "raw"
    (show raw-record matrix)
:param s2_fuzz: extra time around main S2 [ns]
:param s1_fuzz: extra time around main S1 [ns]
:param max_peaks: max peaks for plotting in the wf plot
:param display_peak_info: tuple, items that will be extracted from
    event and displayed in the event info panel see above for format
:param display_event_info: tuple, items that will be extracted from
    event and displayed in the peak info panel see above for format
:param s1_hp_kwargs: dict, optional kwargs for S1 hitpatterns
:param s2_hp_kwargs: dict, optional kwargs for S2 hitpatterns
:param event_time_limit = overrides x-axis limits of event plot
:param plot_all_positions if True, plot best-fit positions
    from all posrec algorithms
"""
event_returns = """
:return: axes used for plotting:
    ax_s1, ax_s2, ax_s1_hp_t, ax_s1_hp_b,
    ax_event_info, ax_peak_info, ax_s2_hp_t, ax_s2_hp_b,
    ax_ev,
    ax_rec
    Where those panels (axes) are:
        - ax_s1, main S1 peak
        - ax_s2, main S2 peak
        - ax_s1_hp_t, S1 top hit pattern
        - ax_s1_hp_b, S1 bottom hit pattern
        - ax_s2_hp_t, S2 top hit pattern
        - ax_s2_hp_b, S2 bottom hit pattern
        - ax_event_info, text info on the event
        - ax_peak_info, text info on the main S1 and S2
        - ax_ev, waveform of the entire event
        - ax_rec, (raw)record matrix (if any otherwise None)
"""

# Add the same docstring to each of these functions
for event_function in (event_display, event_display_simple, _event_display):
    doc = event_function.__doc__
    if doc is not None:
        event_function.__doc__ = doc.format(event_docs=event_docs, event_returns=event_returns)



import numpy as np
import numba
import strax
import straxen

from .peak_ambience import distance_in_xy, _quick_assign


export, __all__ = strax.exporter()


@export
class PeakProximity(strax.OverlapWindowPlugin):
    """Look for peaks around a peak to determine how many peaks are in proximity (in time) of a
    peak."""

    __version__ = "0.6.0"

    depends_on = ("peak_basics", "peak_positions")
    provides = "peak_proximity"
    dtype = [
        # when proximity_exponents is (1, -1, *), the unit is [/ms]
        ("proximity_score", np.float32, "Strength of proximity peaks in (time, space)"),
        (
            "n_competing_left",
            np.int32,
            "Number of larger or slightly smaller peaks left of the main peak",
        ),
        ("n_competing", np.int32, "Number of nearby larger or slightly smaller peaks"),
    ] + strax.time_fields

    proximity_type = straxen.URLConfig(
        default=2,
        type=(int, list),
        help="(List of) type(s) of peaks to consider for proximity score",
    )

    proximity_min_area_fraction = straxen.URLConfig(
        default=0.2,
        infer_type=False,
        help=(
            "The area of competing peaks must be at least "
            "this fraction of that of the considered peak"
        ),
    )

    proximity_window = straxen.URLConfig(
        default=int(1e7),
        infer_type=False,
        help="Peaks starting within this time window (on either side) in ns count as nearby.",
    )

    proximity_exponents = straxen.URLConfig(
        default=(1.0, -1.0, 1.0),
        type=(list, tuple),
        help="The exponent of (delta t, delta r) when calculating proximity score",
    )

    # TODO: change the config name to be more descriptive
    proximity_sigma = straxen.URLConfig(
        default=20.0,
        type=(int, float),
        help="The parameter of HalfCauchy, which is a function of S2 area",
    )

    def get_window_size(self):
        return 10 * self.proximity_window

    def compute(self, peaks):
        argsort = strax.stable_argsort(peaks["center_time"])
        _peaks = peaks[argsort].copy()
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_proximity(peaks, _peaks))
        return result

    def compute_proximity(self, peaks, current_peak):
        result = np.zeros(len(current_peak), dtype=self.dtype)
        strax.set_nan_defaults(result)
        result["time"] = current_peak["time"]
        result["endtime"] = current_peak["endtime"]

        # only consider S1 and S2 peaks
        mask = np.isin(peaks["type"], self.proximity_type)

        roi = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi["time"] = current_peak["center_time"] - self.proximity_window
        roi["endtime"] = current_peak["center_time"].copy()
        result["proximity_score"] = self.peaks_proximity(
            current_peak,
            peaks[mask],
            strax.touching_windows(peaks[mask], roi),
            self.proximity_exponents,
            self.proximity_sigma,
            min_area_fraction=self.proximity_min_area_fraction,
        )

        roi = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi["time"] = current_peak["center_time"] - self.proximity_window
        roi["endtime"] = current_peak["center_time"] + self.proximity_window
        result["n_competing_left"], result["n_competing"] = self.find_n_competing(
            current_peak,
            peaks[mask],
            strax.touching_windows(peaks[mask], roi),
            fraction=self.proximity_min_area_fraction,
        )
        return result

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def find_n_competing(peaks, nearby_peaks, windows, fraction):
        n_left = np.zeros(len(peaks), dtype=np.int32)
        n = n_left.copy()
        nearby_areas = nearby_peaks["area"]
        areas = peaks["area"]

        dig = np.searchsorted(nearby_peaks["center_time"], peaks["center_time"])
        for i, peak in enumerate(peaks):
            left_i, right_i = windows[i]
            threshold = areas[i] * fraction
            n_left[i] = np.sum(nearby_areas[left_i : dig[i]] > threshold)
            n[i] = n_left[i] + np.sum(nearby_areas[dig[i] : right_i] > threshold)

        return n_left, n

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def peaks_proximity(
        peaks,
        pre_peaks,
        touching_windows,
        exponents,
        proximity_sigma,
        min_area_fraction=0.0,
    ):
        sum_array = np.zeros(len(peaks), np.float32)
        for p_i, suspicious_peak in enumerate(peaks):
            if suspicious_peak["area"] <= 0:
                continue
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                creating_peak = pre_peaks[idx]
                dt = suspicious_peak["center_time"] - creating_peak["center_time"]
                dt /= 1e6  # convert from ns to ms
                if (dt <= 0) or (creating_peak["area"] <= 0):
                    continue
                if creating_peak["area"] < min_area_fraction * suspicious_peak["area"]:
                    continue
                dr = distance_in_xy(suspicious_peak, creating_peak)
                if np.isnan(dr):
                    continue
                if dr == 0:
                    # the peaks are at the same position
                    sum_array[p_i] = np.inf
                    continue
                r = creating_peak["area"] / suspicious_peak["area"]
                sigma = proximity_sigma * np.sqrt(
                    1 / creating_peak["area"] + 1 / suspicious_peak["area"]
                )
                pdf = half_cauchy_pdf(dr, sigma).item()
                score = r ** exponents[0] * dt ** exponents[1] * pdf ** exponents[2]
                sum_array[p_i] += score
        return sum_array


@numba.njit
def half_cauchy_pdf(x, scale=1.0):
    return np.where(
        x < 0.0,
        0.0,
        np.where(
            scale <= 0.0,
            np.nan,
            2.0 / (np.pi * scale * (1 + (x / scale) ** 2)),
        ),
    )



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class PeakTopBottomParams(strax.Plugin):
    """Pluging that computes timing characteristics of top and bottom waveforms based on waveforms
    stored at peak level."""

    depends_on = ("peaks", "peak_basics")
    provides = "peak_top_bottom_params"
    __version__ = "0.0.1"

    def infer_dtype(self):
        dtype = []
        peak_basics_fields = self.deps["peak_basics"].dtype.fields
        self.arrs = ["top", "bot"]
        for arr_ in self.arrs:
            dtype += [
                (
                    (f"Central time for {arr_} PMTs [ ns ]", f"center_time_{arr_}"),
                    peak_basics_fields["center_time"][0],
                )
            ]
            dtype += [
                (
                    (
                        f"Time between 10% and 50% area quantiles for {arr_} PMTs [ns]",
                        f"rise_time_{arr_}",
                    ),
                    peak_basics_fields["rise_time"][0],
                )
            ]
            dtype += [
                (
                    (
                        f"Width (in ns) of the central 50% area of the peak for {arr_} PMTs",
                        f"range_50p_area_{arr_}",
                    ),
                    peak_basics_fields["range_50p_area"][0],
                )
            ]
            dtype += [
                (
                    (
                        f"Width (in ns) of the central 90% area of the peak for {arr_} PMTs",
                        f"range_90p_area_{arr_}",
                    ),
                    peak_basics_fields["range_90p_area"][0],
                )
            ]
        dtype += [
            (
                (
                    f"Difference between center times of top and bottom arrays [ ns ]",
                    f"center_time_diff_top_bot",
                ),
                peak_basics_fields[f"center_time"][0],
            )
        ]
        dtype += strax.time_fields
        return dtype

    def compute(self, peaks):
        result = np.zeros(peaks.shape, dtype=self.dtype)
        peak_dtype = self.deps["peaks"].dtype
        for arr_ in self.arrs:
            fpeaks_ = np.zeros(peaks.shape[0], dtype=peak_dtype)
            if arr_ == "top":
                fpeaks_["data"] = peaks["data_top"]
                fpeaks_["area"] = peaks["area"] * peaks[f"area_fraction_top"]
            elif arr_ == "bot":
                fpeaks_["data"] = peaks["data"] - peaks["data_top"]
                fpeaks_["area"] = peaks["area"] * (1.0 - peaks["area_fraction_top"])
            else:
                raise RuntimeError(f"Received unknown array type : " + arr_)
            fpeaks_["length"] = peaks[f"length"]
            fpeaks_["dt"] = peaks[f"dt"]
            mask = fpeaks_["area"] > 0
            # computing center times
            with np.errstate(divide="ignore", invalid="ignore"):
                recalc_ctime = np.sum(
                    fpeaks_["data"] * (np.arange(0, fpeaks_["data"].shape[1])), axis=1
                )
                recalc_ctime /= fpeaks_["area"]
                recalc_ctime *= fpeaks_["dt"]
                recalc_ctime[~mask] = 0.0
            result[f"center_time_{arr_}"] = peaks["time"]
            result[f"center_time_{arr_}"][mask] += recalc_ctime[mask].astype(int)
            # computing widths times
            strax.compute_properties(fpeaks_)
            result[f"rise_time_{arr_}"][:] = np.nan
            result[f"rise_time_{arr_}"][mask] = -fpeaks_["area_decile_from_midpoint"][mask][:, 1]
            result[f"range_50p_area_{arr_}"][:] = np.nan
            result[f"range_50p_area_{arr_}"][mask] = fpeaks_["width"][mask][:, 5]
            result[f"range_90p_area_{arr_}"][:] = np.nan
            result[f"range_90p_area_{arr_}"][mask] = fpeaks_["width"][mask][:, 9]
        result["center_time_diff_top_bot"] = result[f"center_time_top"] - result[f"center_time_bot"]
        result["time"], result["endtime"] = peaks["time"], strax.endtime(peaks)
        return result



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventInfoDouble(strax.MergeOnlyPlugin):
    """Alternate version of event_info for Kr and other double scatter.

    analyses:
     - Uses a different naming convention:
       s1 -> s1_a, alt_s1 -> s1_b, and similarly for s2s;
     - Adds s1_b_distinct_channels, which can be tricky to compute
       (since it requires going back to peaks)

    """

    __version__ = "0.1.2"
    depends_on = ("event_info", "distinct_channels")
    save_when = strax.SaveWhen.EXPLICIT

    @staticmethod
    def rename_field(orig_name):
        special_cases = {
            "cs1": "cs1_a",
            "alt_cs1": "cs1_b",
            "alt_s1_delay": "ds_s1_dt",
            "cs2": "cs2_a",
            "alt_cs2": "cs2_b",
            "alt_s2_delay": "ds_s2_dt",
            "cs1_wo_timecorr": "cs1_a_wo_timecorr",
            "alt_cs1_wo_timecorr": "cs1_b_wo_timecorr",
            "cs2_wo_elifecorr": "cs2_a_wo_elifecorr",
            "alt_cs2_wo_elifecorr": "cs2_b_wo_elifecorr",
            "cs2_wo_timecorr": "cs2_a_wo_timecorr",
            "alt_cs2_wo_timecorr": "cs2_b_wo_timecorr",
            "cs2_area_fraction_top": "cs2_a_area_fraction_top",
            "alt_cs2_area_fraction_top": "cs2_b_area_fraction_top",
            "cs2_bottom": "cs2_a_bottom",
            "alt_cs2_bottom": "cs2_b_bottom",
        }
        if orig_name in special_cases:
            return special_cases[orig_name]

        name = orig_name
        for s_i in [1, 2]:
            if name.startswith(f"s{s_i}"):
                name = name.replace(f"s{s_i}", f"s{s_i}_a")
            if name.startswith(f"alt_s{s_i}"):
                name = name.replace(f"alt_s{s_i}", f"s{s_i}_b")
        return name

    def infer_dtype(self):
        self.input_dtype = strax.unpack_dtype(self.deps["event_info"].dtype) + [
            strax.unpack_dtype(self.deps["distinct_channels"].dtype)[0]
        ]
        return [
            ((comment, self.rename_field(name)), dt) for (comment, name), dt in self.input_dtype
        ]

    def compute(self, events):
        result = np.zeros(len(events), dtype=self.dtype)
        for (_, name), _ in self.input_dtype:
            result[self.rename_field(name)] = events[name]
        return result
